(function(inDesignMode, designModeAction) {
	'use strict';
	ObjC.import('stdlib');

	Number.prototype.toPaddedString = function toPaddedString(len, padding) {
		padding = padding ? padding : "0";
		len = len > 0 ? len : 1;
		var result = this.toString();
		if (result.length >= len)
			return result;
		return Array(len - result.length + 1).join(padding) + result;
	};

	var KMEngine = (function() {
		var _engineApp;

		return {
			getEngineAppName: function() {
				return "Keyboard Maestro Engine";
			},

			getEngineApp: function() {
				if (!_engineApp)
					_engineApp = Application(this.getEngineAppName());
				return _engineApp;
			},

			getVariable: function(name, required) {
				var result = this.getEngineApp().getvariable(name);
				if (!result && required)
					throw Error("Variable '" + name + "' is empty");
				return result;
			},

			setVariable: function(name, value) {
				this.getEngineApp().setvariable(name, {
					to: value
				});
			},
		};
	})();

	var Consts = {
		Actions: {
			Start: "Start",
			Stop: "Stop",
			Resume: "Resume",
			Reset: "Reset",
			Check: "Check"
		},
		ElapsedTimeFormats: {
			HoursMinutesSecondsMilliseconds: "Hours:Minutes:Seconds.Milliseconds",
			HoursMinutesSeconds: "Hours:Minutes:Seconds",
			MinutesSecondsMilliseconds: "Minutes:Seconds.Milliseconds",
			MinutesSeconds: "Minutes:Seconds",
			SecondsMilliseconds: "Seconds.Milliseconds",
			Seconds: "Seconds",
			Milliseconds: "Milliseconds"
		}
	};

	function getPluginParameters() {
		function getPluginParameter(name, required) {
			var envName = "KMPARAM_" + name.replace(/ /g, "_");
			var result;

			if (!inDesignMode) {
				result = $.getenv(envName);
				return result ? result.trim() : "";
			} else {
				var designingParams = {
					KMPARAM_Stopwatch_Variable_Name: "MyStopwatch",
					KMPARAM_Action: designModeAction || Consts.Actions.Start,
					KMPARAM_Elapsed_Time_Format: Consts.ElapsedTimeFormats.SecondsMilliseconds
				};

				result = designingParams[envName];
				if (result === undefined)
					throw Error("Unknown Plugin Parameter Name '" + name + "' while in designing mode");
			}

			if (required && !result)
				throw Error("'" + name + "' is required");
			return result;
		}

		// getPluginParameters
		return {
			stopwatchVariableName: getPluginParameter("Stopwatch Variable", true),
			action: getPluginParameter("Action", true),
			elapsedTimeFormat: getPluginParameter("Elapsed Time Format", true)
		};
	} // getPluginParameters

	function validateParams(params) {
		function objectValues(o) {
			var keys = Object.keys(o);
			return keys.map(function(key) {
				return o[key];
			});
		}

		function validateValue(name, values, value) {
			if (!value)
				throw Error("'" + name + "' can't be empty");
			var validValues = objectValues(values);
			if (!validValues.includes(value))
				throw Error("'" + name + "' has an invalid value: '" + value + "'. Valid values are: '" + validValues.join("', '") + "'");
		}

		function validateVariableName(name, varName, isRequired) {
			if (!varName) {
				if (isRequired)
					throw Error("'" + name + "' can't be empty");
				return;
			}
			if (varName.search(/^[a-z][a-z0-9 _]*$/gi) < 0)
				throw Error("'" + name + "' does not contain a valid variable name: '" + varName + "'");
		}

		// validateParams
		validateVariableName("Stopwatch Variable", params.stopwatchVariableName, true);
		validateValue("Action", Consts.Actions, params.action);
		validateValue("Elapsed Time Format", Consts.ElapsedTimeFormats, params.elapsedTimeFormat);
	} // validateParams

	var Stopwatch = (function(values, isJSON, now) {
		var _values = {
			id: "Stopwatch.v1.0",
			startTime: now || Date.now(),
			elapsedTime: 0
		};

		if (values)
			_setValues(values, isJSON);

		function _setValues(values, isJSON) {
			if (isJSON) {
				try {
					values = JSON.parse(values);
				} catch (e) {
					throw Error("Invalid Stopwatch 'Values' object");
				}
			}

			if (values.id !== _values.id)
				throw Error("Invalid Stopwatch 'Values' object");
			_values = values;
		}

		return {
			getElapsedTime: function(now) {
				if (_values.startTime > 0)
					return ((now || Date.now()) - _values.startTime) + _values.elapsedTime;
				return _values.elapsedTime;
			},

			getValues: function() {
				return _values;
			},

			isRunning: function() {
				return _values.startTime > 0;
			},

			reset: function() {
				_values.startTime = 0;
				_values.elapsedTime = 0;
			},

			start: function(now) {
				_values.startTime = now || Date.now();
			},

			stop: function(now) {
				if (_values.startTime)
					_values.elapsedTime = _values.elapsedTime + ((now || Date.now()) - _values.startTime);
				_values.startTime = 0;
			}
		};
	});

	function getStopwatch(params) {
		var stopwatch = null;
		try {
			stopwatch = new Stopwatch(KMEngine.getVariable(params.stopwatchVariableName), true);
		} catch (e) {
		}
		if (!stopwatch)
			stopwatch = new Stopwatch();
		return stopwatch;
	}

	function formatElapsedTime(value, format) {
		var mSecond = 1000;
		var mMinute = mSecond * 60;
		var mHour = mMinute * 60;
		var hasHours = format.indexOf("Hours") >= 0;
		var hasMinutes = format.indexOf("Minutes") >= 0;
		var hasSeconds = format.indexOf("Seconds") >= 0;
		var hasMilliseconds = format.indexOf("Milliseconds") >= 0;

		var result = "";
		if (hasHours) {
			var hours = hasMinutes ? Math.floor(value / mHour) : Math.round(value / mHour);
			value = value - hours * mHour;
			result = result + (result ? ":" + hours.toPaddedString(2) : hours);
		}
		if (hasMinutes) {
			var minutes = hasSeconds ? Math.floor(value / mMinute) : Math.round(value / mMinute);
			value = value - minutes * mMinute;
			result = result + (result ? ":" + minutes.toPaddedString(2) : minutes);
		}
		if (hasSeconds) {
			var seconds = hasMilliseconds ? Math.floor(value / mSecond) : Math.round(value / mSecond);
			value = value - seconds * mSecond;
			result = result + (result ? ":" + seconds.toPaddedString(2) : seconds);
		}
		if (hasMilliseconds) {
			result = result + (result ? "." + value.toPaddedString(3) : value);
		}
		return result;
	}

	function doAction(stopwatch, params) {
		var elapsedTime;
		switch (params.action) {
			case Consts.Actions.Start:
				stopwatch.reset();
				stopwatch.start();
				elapsedTime = 0;
				break;
			case Consts.Actions.Stop:
				stopwatch.stop();
				elapsedTime = stopwatch.getElapsedTime();
				break;
			case Consts.Actions.Resume:
				elapsedTime = stopwatch.getElapsedTime();
				if (!stopwatch.isRunning())
					stopwatch.start();
				break
			case Consts.Actions.Reset:
				elapsedTime = stopwatch.getElapsedTime();
				stopwatch.reset();
				break;
			case Consts.Actions.Check:
				elapsedTime = stopwatch.getElapsedTime();
				break;
			default:
				throw Error("Unexpected Action: '" + params.action + "'");
		}

		KMEngine.setVariable(params.stopwatchVariableName, JSON.stringify(stopwatch.getValues()));

		return formatElapsedTime(elapsedTime, params.elapsedTimeFormat);
	}

	function execute() {
		var params = getPluginParameters();
		validateParams(params);
		var stopwatch = getStopwatch(params);
		return doAction(stopwatch, params);
	}

    if (inDesignMode) {
        return execute();
    } else {
        try {
            return execute();
        } catch (e) {
            return "Error: " + e.message;
        }
    }
})(true, "Stop");
